/*
 * dParse
 * D module parser for leds
 * 
 * by Antonio Monteiro
 *
 * This is actually an independent project
 * that was merged into leds to facilitate
 * distribution.
 */

module parse.Module;

private import std.c.stdio; // note
private import std.string;
private import std.path;

alias std.path filePath; 

private import std.file; 


enum TypeOfElement
{
	NONE,
	CONST, 
	VAR,
	CONSTRUCTOR,
	FUNCTION,
	ENUM,
	STRUCT,
	INTERFACE, 
	ABSCLASS,
	CLASS,
	MODULE
}

enum Access
{
	NONE,
	PUBLIC,
	PROTECTED,
	PRIVATE
}

struct Attributes
{
	bit	attConst;
	bit	attAbstract;
	bit	attStatic;
	bit	attFinal;
}

class Parser
{
	private Access currentAccess = Access.NONE;
	private char[] input;
	char[][] tokens;
	private char[] token;
	private int mainPos;
	private int tPos;

	this(char[] input)
	{
		mainPos = 0;
		tPos = 0;
		this.input = input;
	}

	/**
	 * parses the access protection.
	 * any element can be preceded with a access protection
	 */
	Access getAccess(char[][] tokens, inout int tPos)
	{
		bit setCurrent = false;
		Access tempAccess = Access.NONE;
		if ( tPos < tokens.length )
		{
			char[] token = tokens[tPos++];
			switch ( token )
			{
				case "public":	tempAccess = Access.PUBLIC;	break;
				case "protected":
					tempAccess = Access.PROTECTED;
					break;
					
				case "private":
					tempAccess = Access.PRIVATE;
					break;		
				default:
					tempAccess = Access.NONE;
					--tPos;
					break;
			}
			if ( tPos<tokens.length && tokens[tPos] == ":" )
			{
				setCurrent = true;
				++tPos;
			}
		}
		if ( setCurrent && tempAccess != Access.NONE )
		{
			currentAccess = tempAccess;
		}
		return tempAccess == Access.NONE ? currentAccess : tempAccess;
	}

	/**
	 * any element can be preceded by zero or more attributes
	 */
	Attributes getAttributes(char[][] tokens, inout int tPos)
	{
		Attributes att;
		
		bit endAtt = false;
		
		while ( tPos < tokens.length && !endAtt)
		{
			char[] token = tokens[tPos++];
			endAtt = false;
			switch ( token )
			{
				case "const": att.attConst = true; break;
				case "abstract": att.attAbstract = true; break;
				case "static": att.attStatic = true; break;
				case "final": att.attFinal = true; break;
				default:
					endAtt = true;
					tPos--;
					break;
			}
		}
		return att;
	}

	/**
	 * Every element must have a type and a name.
	 * type is the actual type of the value (int, float, class...)
	 */
	char[] getTypeAndName(char[][] tokens, inout int tPos, inout char[] name)
	{
		char[] type;
		if ( tPos < tokens.length )
		{
			char[] token = tokens[tPos++];
			if ( token in knownTypes )
			{
				type =  token.dup;
				
				if ( tPos < tokens.length )
				{
					bit arrayDefinition = true;
					int count = 0;
					while( tPos<tokens.length && arrayDefinition)
					{
						if ( tokens[tPos] == "[" || tokens[tPos]== "]" )
						{
							if ( count++ == 0 )
							{
								type ~= ' ';
							}
							type ~= tokens[tPos++][0];
						}
						else
						{
							arrayDefinition = false;
						}
					}
				
					if ( tokens[tPos] != "{" )
					{
						name = tokens[tPos++].dup;
					}
				}
			}
			else
			{
				if ( tPos < tokens.length )
				{
					bit validTypeDefinition = false;
					bit constructorDefinition = false;
					name = "";
					//token = tokens[tPos++];
					switch ( token )
					{
						case "alias":
						case "typedef":
							if ( tPos < tokens.length )
							{
								token = tokens[tPos++];
							}
							else
							{
								break;
							}
						case "class":
						case "interface":
						case "enum":
						case "struct":
						case "union":
							type = token.dup;
							validTypeDefinition = true;
							break;
							
						case "this":
							type = "ctor";
							name = token.dup;
							constructorDefinition = true;
							break;
							
						default:
							validTypeDefinition = false;
							break;

					}
					if ( validTypeDefinition )
					{
						if ( tPos < tokens.length )
						{
							if ( tokens[tPos] != "{" )
							{
								name = tokens[tPos++];
								knownTypes[name] = token;
							}
							else
							{
								knownTypes[token~"."~name] = token;
							}
						}
					}
					else if ( constructorDefinition )
					{
						// nothing here
					}
					else
					{
						type = token.dup;
						bit arrayDefinition = true;
						int count = 0;
						while( tPos<tokens.length && arrayDefinition)
						{
							if ( tokens[tPos] == "[" || tokens[tPos]== "]" )
							{
								if ( count++ == 0 )
								{
									type ~= ' ';
								}
								type ~= tokens[tPos++][0];
							}
							else
							{
								arrayDefinition = false;
							}
						}
						if ( tPos < tokens.length )
						{
							if ( tokens[tPos] != "{" )
							{
								name = tokens[tPos++].dup;
							}
						}
					}
					
				}
			}
		}
		return type;
	}

	/**
	 * skips until the char enSkip
	 * @param endSkip the char that marks the end of the skip
	 * \todo parse comments and literals out
	 */
	bit skip(char endSkip)
	{
		if ( mainPos < input.length )
		{
			char c = input[mainPos];
			while ( mainPos<input.length && c!=endSkip)
			{
				++mainPos;
				c = input[mainPos];
			}
		}
		return mainPos < input.length;
	}
	
	bit skipFunctionBody()
	{
		if ( mainPos < input.length )
		{
			char c = ' ';
			int level = 1;
			bit inString = false;
			bit inChar = false;
			bit scape = false;
			char prevC = ' ';
			while ( mainPos<input.length && level > 0)
			{
				c = input[mainPos++];
				
				if ( c == '\'' )
				{
					prevC = '\\';
					while ( mainPos<input.length
							&& (c!='\'' || prevC == '\\') )
					{
						if ( prevC == '\\' )
						{
							prevC = ' ';
						}
						else
						{
							prevC = c;
						}
						c = input[mainPos++];
					}
				}
				else if ( c == '"' )
				{
					prevC = '\\';
					while ( mainPos<input.length
							&& ( c!='"' || prevC == '\\' ) )
					{
						if ( prevC == '\\' )
						{
							prevC = ' ';
						}
						else
						{
							prevC = c;
						}
						c = input[mainPos++];
					}
				}
				
				if ( c == '{' )
				{
					++level;
				}
				else if ( c == '}' )
				{
					--level;
				}
				
			}
			if ( mainPos < level )
			{
				--mainPos;
			}
		}
		return mainPos < input.length;
	}

	Element getElement()
	{
		Element element;
		int tPos = 0;
		if ( nextTokens('}') )
		{
			//printf("getElement mainPos = %d\n",mainPos);
			Access tempAccess = getAccess(tokens, tPos);
			Attributes att = getAttributes(tokens, tPos);
			char[] name;
			char[] type = getTypeAndName(tokens, tPos, name);
			
			bit typeFunction = false;
			if ( tPos<tokens.length && tokens[tPos]=="(" )
			{
				typeFunction = true;
				++tPos;
				element = new Function(type, name, tempAccess, this);
			}
			else
			{
				switch(type)
				{
					case "enum":
						element = new Enum(type, name, tempAccess, this);
						break;
						
					case "class":
						element = new Class(name, tempAccess, this);
						break;
						
					default:
						element = new Element(type, name, tempAccess);
						break;
				}
			}
			
			return element;
		}
		else
		{
			return null;
		}
	}

	/**
	 * gets the tokens until a ";", a "{" or a "(" or a received char
	 * \todo this is a good opportunity to practice regexps
	 */
	bit nextTokens(char breakChar)
	{
		char[] statment;
		bit inString = false;
		if ( mainPos < input.length )
		{
			char c = input[mainPos];
			char[] comment;
			while ( mainPos<input.length && c!=';' && c!='{' && c!='(' && c!=breakChar)
			{
				++mainPos;
				if ( c == '/' && statment.length>0 && statment[statment.length-1]=='/' )
				{
					skipEOLComment();
					statment.length = statment.length - 1;
				}
				else if ( (c=='+' || c=='*') && statment.length>0 && statment[statment.length-1]=='/' )
				{
					skipMultiComment(c);
					statment.length = statment.length - 1;
				}
				else if ( .find("()[]:,.",c) >= 0 )
				{
					statment ~= ' ';
					statment ~= c;
					statment ~= ' ';
				}
				else
				{
					statment ~= c;
				}
				if ( mainPos<input.length )
				{
					c = input[mainPos];
				}
				
			}
			if ( .find(";{(",c)>=0 || c==breakChar)
			{
				++mainPos;
				statment ~= ' ';
				statment ~= c;
			}
			//printf("statment >>>>>>>>>>>> %.*s\n<<<<<<<<<\n",statment);
			tokens = .split(strip(statment));
			return true;
		}
		return false;
	}

	/**
	 * skips a single line comment
	 */	
	void skipEOLComment()
	{
		while( mainPos<input.length && input[mainPos]!='\n' )
		{
			++mainPos;
		}
		
	}
	
	/**
	 * skips a multi line comment but returns it
	 */
	char[] skipMultiComment(char oC)
	{
		char[] comment;
		bit keepComment = true;
		char prev = ' ';
		bit end = false;
		++mainPos;
		if ( mainPos<input.length )
		{
			char c = input[mainPos];
			if ( oC=='*' && c=='*' )
			{
				keepComment = true;
			}
			do
			{
				prev = c;
				c = input[mainPos++];
				if ( keepComment )
				{
					comment ~= c;
				}
			} while( mainPos<input.length && (prev!='*' || c!='/') )
		}
		
		return keepComment ? comment : cast(char[])null;
	}

}
/**
 * Contains the basic information for the simple types.
 * Also contains the generic parsing functions
 */
class Element
{

	Access access;
	char[] id;
	TypeOfElement typeOfElement;
	char[] returnType;
	char[] comment;

	Element parent;

	bit array;

	bit attConst;
	bit attAbstract;
	bit attStatic;
	bit attFinal;

	Access tempAccess = Access.NONE;
	
	char[] accessStr(Access access)
	{
		char[] str = "";
		switch (access)
		{
			case Access.NONE: str = ""; break;
			case Access.PUBLIC: str = "public "; break;
			case Access.PROTECTED: str = "protected "; break;
			case Access.PRIVATE: str = "private "; break;
			default: str = ""; break;
		}
		return str;
	}
	
	/** when true this element has ended */
	bit elementEnd = false;
	
	this(char[] returnType, char[] name, Access access)
	{
		//printf("new Element %.*s %.*s\n",returnType, name);
		id = name.dup;
		this.access = access;
		this.returnType = returnType.dup;
	}

	void setParent(Element element)
	{
		parent = element;
	}
	
	char[] getParentName()
	{
		char[] parentName;
		
		if ( parent !== null )
		{
			parentName = parent.getParentName() ~ "." ~ parent.id;
		}
		
		return parentName;
	}
	
	void print()
	{
		printf("Element %.*s%.*s %.*s\n",accessStr(access),returnType, id);
	}

	char[] getLongForm()
	{
		return accessStr(access)~" "~returnType~" "~id;
	}

	/**
	 * adds an element to this element
	 */
	void addElement(Element element)
	{
		// do nothing - this basic element cannot have elements
		printf("ERROR - adding element to basic element\n");
	}

}

class Enum : Element
{
	char[][] entries;
	
	this(char[] returnType, char[] name, Access access)
	{
		super(returnType, name, access);
	}

	this(char[] returnType, char[] name, Access access, Parser parser)
	{
		super(returnType, name, access);
		parser.nextTokens('}');
		addEntries(parser.tokens);
	}

	void addEntries(char[][] tokens)
	{
		int i = 0;
		while ( i < tokens.length )
		{
			entries ~= tokens[i++];
			while ( (i<tokens.length) && tokens[i]!="," )
			{
				++i;
			}
			++i;
		}
	}

	void print()
	{
		print(entries);
	}

	void print(char[][] entries)
	{
		printf("Enum : %.*s%.*s {\n",accessStr(access),id);
		foreach(char[] entry ; entries)
		{
			printf("  %.*s\n",entry);
		}
		printf("} // enum %.*s\n",id);
	}
}

/**
 * contains information on a function element
 */
class Function : Element
{
	char[][] modifier;
	char[][] types;
	char[][] parameters;
	
	this(char[] returnType, char[] name, Access access)
	{
		super(returnType, name, access);
	}
	
	this(char[] returnType, char[] name, Access access, Parser parser)
	{
		super(returnType, name, access);
		getParameters(parser);
		parser.skipFunctionBody();
	}
	
	void getParameters(Parser parser)
	{
		parser.nextTokens('}');
		char[][] tokens = parser.tokens;
		int i = 0;
		char[] token;

		if ( tokens[tokens.length-1] == "{" )
		{
			tokens.length = tokens.length-1;
		}
		if ( tokens[tokens.length-1] == ")" )
		{
			tokens.length = tokens.length-1;
		}

		enum Getting { MOD, TYPE, NAME };
		Getting getting = Getting.MOD;

		while ( i < tokens.length )
		{
			token = tokens[i++];
			switch ( getting )
			{
				case Getting.MOD:
					if ( token == "," )
					{
						if ( i<tokens.length )
						{
							token = tokens[i++];
						}
					}
					if ( token=="in" || token=="out" || token=="inout" )
					{
						modifier ~= token.dup;
					}
					else
					{
						modifier ~= "";
						--i;
					}
					getting = Getting.TYPE;
					break;
					
				case Getting.TYPE:
					char[] type = token;
					bit arrayDefinition = true;
					int count = 0;
					while( i<tokens.length && arrayDefinition)
					{
						if ( tokens[i] == "[" || tokens[i] == "]" )
						{
							if ( count++ == 0 )
							{
								type ~= ' ';
							}
							type ~= tokens[i++][0];
						}
						else
						{
							arrayDefinition = false;
						}
					}
					types ~= type.dup;
					getting = Getting.NAME;
					break;
					
				case Getting.NAME:
					char[] name = token;
					while ( i<tokens.length && tokens[i] != ",")
					{
						name = name ~ tokens[i++];
					}
					parameters ~= name.dup;
					getting = Getting.MOD;
					break;
			}


		}
	}

	void print()
	{
		printf("Function %.*s%.*s %.*s(",accessStr(access),returnType, id);
		bit second = false;
		for ( int i=0 ; i<types.length && i<parameters.length ; i++ )
		{
			if ( second )
			{
				printf(", ");
			}
			else
			{
				second=true;
			}
			if ( modifier[i].length > 0 )
			{
				printf("%.*s",modifier[i]);
			}
			printf("%.*s %.*s",types[i], parameters[i]);
		}
		printf(")\n");
	}

	char[] getLongForm()
	{
		char[] str = accessStr(access)~" "~returnType~" "~id~"(";
		bit second = false;
		for ( int i=0 ; i<types.length && i<parameters.length ; i++ )
		{
			if ( second )
			{
				str ~= ", ";
			}
			else
			{
				second=true;
			}
			if ( modifier[i].length > 0 )
			{
				str ~= modifier[i];
			}
			str ~= types[i]~" "~parameters[i];
		}
		str ~= ")";
		return str;
	}


}

/**
 * Contains information on a class element
 */
class Class : Element
{
	/**
	 * elements of a class are anything but functions (methods)
	 * and others nested classes
	 */
	Element[] elements;
	
	/**
	 * all the enums of this class
	 */
	Enum[] enums;
	/**
	 * functions of a class are the methods
	 */
	Function[] functions;
	/**
	 * classes of a class are the nested classes
	 */
	Class[] classes;
	
	/**
	 * modules are the modules imported inside the class
	 */
	Module[] modules;

	this(char[] name, Access access)
	{
		super(name, name, access);
	}

	this(char[] name, Access access, Parser parser)
	{
		super(name, name, access);
		parse(parser);
	}

	/**
	 * Recursive to parse tokens (I had the good idea but...).
	 * we get the class of object that contains the tokens
	 */
	void parse(Parser parser)
	{
		char[] comment;
		Element element = parser.getElement();
		//printf("start class parse ------------------- \n");
		while ( !elementEnd && element )
		{
			addElement(element);
			elementEnd = parser.tokens.length == 1 && parser.tokens[0] == "}";
			if ( ! elementEnd )
			{
				element = parser.getElement();
			}
		}
		//printf("end class parse ------------------- \n");
		elementEnd = false;
	}

	/**
	 * Adds an element to this Module.
	 * A modules can contain any type of element
	 */
	void addElement(Element element)
	{
		if ( element !== null )
		{
			element.setParent(this);
			//printf("addElement %.*s %.*s\n",element.returnType, element.id);
			if ( cast(Module)element )
			{
				//printf("    addElement is cast to Module\n");
				modules ~= cast(Module)element;
			}
			else if ( cast(Class)element )
			{
				//printf("    addElement is cast to Class\n");
				classes ~= cast(Class)element;
			}
			else if ( cast(Function)element )
			{
				//printf("    addElement is cast to Function\n");
				functions ~= cast(Function)element;
			}
			else if ( cast(Enum) element )
			{
				//printf("    addElement is cast to Enum\n");
				enums ~= cast(Enum)element;
			}
			else
			{
				//printf("    addElement is cast to element\n");
				elements ~= element;
			}
		}
		
	}

	void print()
	{
		printf("Class %.*s%.*s ##########\n",accessStr(access),id);
		printElements();
	}

	void printElements()
	{
		foreach(Element element; elements)
		{
			element.print();
		}
		foreach(Enum element; enums)
		{
			element.print();
		}
		foreach(Function element; functions)
		{
			element.print();
		}
		foreach(Class element; classes)
		{
			element.print();
		}
		foreach(Module element; modules)
		{
			element.print();
		}
	}

}

private char[][char[]] knownTypes;

/**
 * A class to contain a module structure of elements
 */
public:
class Module : Class
{

	char[] fileName;
	char[] moduleName;

	this(char[] filename)
	{
		//printf("parseing module %.*s\n",filename);
		super(filename, Access.NONE);
		this.fileName = filename.dup;
		typeOfElement = TypeOfElement.MODULE;
		setupTypes();
		char[] input = loadFile(fileName);
		Parser parser = new Parser(input);
		parse(parser);
	}

	this(char[] text, Access access)
	{
	//printf("Module.this %.*s\n",text[0..240]);
		super("", access);
		this.fileName = "";
		typeOfElement = TypeOfElement.MODULE;
		setupTypes();
		Parser parser = new Parser(text);
		parse(parser);
	}
	
	/**
	 * Recursive to parse tokens (I had the good idea but...).
	 * we get the class of object that contains the tokens
	 */
	void parse(Parser parser)
	{
		char[] comment;
		Element element = parser.getElement();
		//printf("start class parse ------------------- \n");
		while ( !elementEnd && element )
		{
			addElement(element);
			//elementEnd = parser.tokens.length == 1 && parser.tokens[0] == "}";
			if ( ! elementEnd )
			{
				element = parser.getElement();
			}
		}
		//printf("end class parse ------------------- \n");
		elementEnd = false;
	}

	void print()
	{
		printf("Module %.*s%.*s ########## ##########\n",accessStr(access),id);
		printElements();
		//printTypes();
	}
	
	void printClasses(bit showClassElements)
	{
	
		foreach(Class element ; classes)
		{
			printf("######### >>> Module Class %.*s\n",element.getLongForm());
			if ( showClassElements )
			{
				element.printElements();
			}
		}
	}

	void printElements()
	{
		foreach(Element element ; elements)
		{
			printf("######### >>> Module Element %.*s\n",element.getLongForm());
		}
	}

	void printTypes()
	{
		foreach(char[] key ; knownTypes.keys)
		{
			printf("type = %.*s\n",key);
		}
	}

	void setupTypes()
	{
		knownTypes["void"] = "no type";
		knownTypes["bit"] = "single bit";
		knownTypes["byte"] = "signed 8 bits";
		knownTypes["ubyte"] = "unsigned 8 bits";
		knownTypes["short"] = "signed 16 bits";
		knownTypes["ushort"] = "unsigned 16 bits";
		knownTypes["int"] = "signed 32 bits";
		knownTypes["uint"] = "unsigned 32 bits";
		knownTypes["long"] = "signed 64 bits";
		knownTypes["ulong"] = "unsigned 64 bits";
		knownTypes["cent"] = "signed 128 bits (reserved for future use)";
		knownTypes["ucent"] = "unsigned 128 bits (reserved for future use)";
		knownTypes["float"] = "32 bit floating point";
		knownTypes["double"] = "64 bit floating point";
		knownTypes["real"] = "largest hardware implemented floating point size (Implementation Note: 80 bits for Intel CPU's)";
		knownTypes["ireal"] = "a floating point value with imaginary type";
		knownTypes["ifloat"] = "imaginary float";
		knownTypes["idouble"] = "imaginary double";
		knownTypes["creal"] = "a complex number of two floating point values";
		knownTypes["cfloat"] = "complex float";
		knownTypes["cdouble"] = "complex double";
		knownTypes["char"] = "unsigned 8 bit UTF-8";
		knownTypes["wchar"] = "unsigned 16 bit UTF-16";
		knownTypes["dchar"] = "unsigned 32 bit UTF-32";

	}

	/**
	 * loads a file into the view
	 * @param fileName the file to load
	 * @return true if sucesseful
	 */
	char[] loadFile(char[] fileName)
	{
		char[] input;
		this.fileName = fileName.dup;
		try
		{
			input = cast(char[])std.file.read(fileName);

		}
		catch ( Error e)
		{
			e.print();
			input = null;
		}
		return input;
	}

}

//version(lib)
//{
//}
//else
//{
//void main(char[][] args)
//{
//
//	Module m;
//	
//	if ( args.length > 1 )
//	{
//		m = new Module(args[1]);
//	}
//	else
//	{
//		m = new Module("/home/ruimt/devel/D/dParse/src/parse/Module.d");
//	}
//	printf("########### these are the module classes >>>>>>>>>>>>>>\n");
//	m.printClasses(false);
//	printf("<<<<<<<<<<< these are the module classes ##############\n");
//	m.printElements();
//	m.print();
//}
//}
